#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2021 Surface Concept GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

-------------------------------------------------------------------------------

Created on Thu Jun 15 16:29:00 2021

WARNING: For simplicity and brevity of code this example does not show the
best way of handling the received data. Make sure to have a look at
example_buffered_data_callbacks3.py. This example is still useful for some
of the comments and demonstrating the basic principle.

Test of the buffered-data callbacks interface.
As of June 2021, this interface should be the method of choice if event data
is to be processed in python code. The older USER_CALLBACKS pipe should not
be used from Python anymore (but is still available for backwards
compatibility).

"""

import scTDC
import numpy as np

# -----------------------------------------------------------------------------
# example 1 of deriving from buffered_data_callbacks_pipe

# define a bitmask which selects the parts of the event data that will be
# buffered
DATA_FIELD_SEL1 = \
      scTDC.SC_DATA_FIELD_DIF1 \
    | scTDC.SC_DATA_FIELD_DIF2 \
    | scTDC.SC_DATA_FIELD_TIME

class BufDataCB1(scTDC.buffered_data_callbacks_pipe):
    def __init__(self, lib, dev_desc,
                 data_field_selection=DATA_FIELD_SEL1,
                 max_buffered_data_len=100000,
                 dld_events=True):
        super().__init__(lib, dev_desc, data_field_selection,  # <-- mandatory!
                         max_buffered_data_len, dld_events)    # <-- mandatory!

    def on_data(self, d):
        # This function is called from the scTDC1 library when (1) the
        # data buffer is full, or (2) on request at the end of a measurement.
        # Being called from the scTDC1 library, it operates within a thread of
        # the scTDC1 library and pauses internal processing of the TDC data
        # stream on the PC until we return from this function.
        # All numpy arrays in the argument 'd' will only be valid until we
        # return from this function! If you want to retain data for longer,
        # you will need numpy.ndarray.copy.
        # See example_buffered_data_callbacks3.py
        print("data callback with {} events".format(d["data_len"]))
        if d["data_len"] == 0:
          return
        print(" statistics x - coordinate")
        statx = (np.mean(d["dif1"]), np.min(d["dif1"]), np.max(d["dif1"]))
        print(" mean {}  min {}  max {}".format(*statx))
        print(" statistics y - coordinate")
        staty = (np.mean(d["dif2"]), np.min(d["dif2"]), np.max(d["dif2"]))
        print(" mean {}  min {}  max {}".format(*staty))
        print(" statistics time values")
        statt = (np.mean(d["time"]), np.min(d["time"]), np.max(d["time"]))
        print(" mean {}  min {}  max {}".format(*statt))

    def on_end_of_meas(self):
        print("\nend of measurement")
        # returning a boolean value of True means that we request buffered
        # data immediately. This will cause an invocation of 'on_data' right
        # after returning from this function, even if the buffer is only half
        # filled
        return True

# -----------------------------------------------------------------------------


def test1():
    device = scTDC.Device(autoinit=False)

    # initialize TDC --- and check for error!
    retcode, errmsg = device.initialize()
    if retcode < 0:
        print("error during init:", retcode, errmsg)
        return 0
    else:
        print("successfully initialized")

    # open a BUFFERED_DATA_CALLBACKS pipe
    bufdatacb1 = BufDataCB1(device.lib, device.dev_desc)

    # do a measurement
    retcode = bufdatacb1.start_measurement_sync(100)
    if retcode < 0:
        print(device.lib.sc_get_err_msg(retcode))

    # clean up
    bufdatacb1.close() # closes the user callbacks pipe, method inherited from base class
    device.deinitialize()


if __name__ == "__main__":
    test1()
