#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2021 Surface Concept GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

-------------------------------------------------------------------------------

Created on Wed Jun 16 10:30:00 2021


Test of the buffered-data callbacks interface.
Similar to example_buffered_data_callbacks2.py, this example performs the
computation of a histogram from event data in python code rather than letting
the scTDC1 library do it.
In contrast to the previous example, *** the computation is delegated to
the main thread. *** This results in better performance: When inside the
on_data(...) callback, the internal data processing of the scTDC library is
paused. Creating a copy of the data and then immediately returning from the
callback spends as little time as possible while still making the data
available for processing in another thread.
Conceptual ideas:
* need the main thread free for doing computations
  -> start measurements asynchronously (i.e. in the background) instead
  of idle-waiting for the measurement to finish.
* Use a queue (good for thread-safe messaging) to pass over data and
  other events from callbacks to main thread. The queue also provides the
  mechanism to let the main thread wait until it gets something to do.
* Make a ("deep") copy of the data in the callback since it will be invalid
  as soon as the callback function returns.
"""

import scTDC
import time
import timeit
import numpy as np
from queue import Queue
from matplotlib import pyplot as plt
#

# -----------------------------------------------------------------------------
# example 3 of deriving from buffered_data_callbacks_pipe

NR_OF_MEASUREMENTS = 2

DATA_FIELD_SEL1 = \
      scTDC.SC_DATA_FIELD_DIF1 \
    | scTDC.SC_DATA_FIELD_DIF2 \
    | scTDC.SC_DATA_FIELD_TIME

# define some constants to distinguish the type of element placed in the queue
QUEUE_DATA = 0
QUEUE_ENDOFMEAS = 1
QUEUE_ENDOFSEQUENCE = 2

class BufDataCB3(scTDC.buffered_data_callbacks_pipe):
    def __init__(self, lib, dev_desc,
                 data_field_selection=DATA_FIELD_SEL1,
                 max_buffered_data_len=500000,
                 dld_events=True):
        super().__init__(lib, dev_desc, data_field_selection,  # <-- mandatory!
                         max_buffered_data_len, dld_events)    # <-- mandatory!

        self.buffer_meas = NR_OF_MEASUREMENTS - 1
        self.queue = Queue()

    def on_data(self, d):
        # make a dict that contains copies of numpy arrays in d ("deep copy")
        # start with an empty dict, insert basic values by simple assignment,
        # insert numpy arrays using the copy method of the source array
        dcopy = {}
        for k in d.keys():
            if isinstance(d[k], np.ndarray):
                dcopy[k] = d[k].copy()
            else:
                dcopy[k] = d[k]
        self.queue.put((QUEUE_DATA, dcopy))

    def on_end_of_meas(self):
        if self.buffer_meas > 0:
            self.buffer_meas -= 1 # count down
            self.queue.put((QUEUE_ENDOFMEAS, None))
            return False          # don't trigger an on_data callback yet
        else:
            self.buffer_meas = NR_OF_MEASUREMENTS
            self.queue.put((QUEUE_ENDOFSEQUENCE, None))
            return True # last measurement, trigger on_data callback

# -----------------------------------------------------------------------------

class DataProcessor():
    """ This class is created to handle the data processing. Using a class for
    such a task has the advantage that we can maintain a state across multiple
    hand-overs of data chunks and the variables holding this state are visibly
    grouped together in the class."""
    def __init__(self):
        self.xy_histo = None
        self.xmax = None
        self.ymax = None
        self.total_events = 0

    def process_data(self, d):
        """ d is the same data structure as in BufDataCB3.on_data(...)
            i.e. a dictionary containing 1D arrays (and some integer values)
        """
        self.total_events += d["data_len"]
        print("data callback with {} events".format(d["data_len"]))
        print(" and {} start of measurements".format(len(d["som_indices"])))
        if d["data_len"] == 0:
          return        
        # demonstrate that arbitrary transforms of the event data are possible
        # rotate the xy image by an angle around a center
        angle = 22.5
        binning = 4
        if self.xmax is None:
            self.xmax = np.max(d["dif1"])
            self.ymax = np.max(d["dif2"])
        imgdim = max(self.xmax, self.ymax)
        xcenter = self.xmax/2.0
        ycenter = self.ymax/2.0
        cosa = np.cos(angle / 180.0 * np.pi)
        sina = np.sin(angle / 180.0 * np.pi)
        xrot = cosa * (d["dif1"] - xcenter) - sina * (d["dif2"] - ycenter) + xcenter
        yrot = sina * (d["dif1"] - xcenter) + cosa * (d["dif2"] - ycenter) + ycenter
        # now build a 2D histogram of it (i.e. an image of the counts vs
        # detector position)
        # note: there is a python module named fast_histogram
        # (which you can install by calling python with the arguments
        #  "-m pip install --user fast_histogram")
        # fast_histogram is supposedly considerably faster than numpy's routine
        h,_xe,_ye = np.histogram2d(xrot, yrot, range=[[0,imgdim], [0,imgdim]],
                           bins=imgdim/binning)
        # (numpy.histogram2d returns a triple holding the result in element 0)
        # cache this histogram data and add histogram data of subsequent invocations
        if self.xy_histo is None:
            self.xy_histo = h
        else:
            self.xy_histo += h

def test3():
    device = scTDC.Device(autoinit=False)
    dataprocessor = DataProcessor()

    # initialize TDC --- and check for error!
    retcode, errmsg = device.initialize()
    if retcode < 0:
        print("error during init:", retcode, errmsg)
        return -1
    else:
        print("successfully initialized")

    # open a BUFFERED_DATA_CALLBACKS pipe
    bufdatacb = BufDataCB3(device.lib, device.dev_desc)

    # define a closure that checks return codes for errors and does clean up
    def errorcheck(retcode):
        if retcode < 0:
            print(device.lib.sc_get_err_msg(retcode))
            bufdatacb.close()
            device.deinitialize()
            return -1
        else:
            return 0

    start = timeit.default_timer()

    # start a first measurement
    retcode = bufdatacb.start_measurement(100)
    if errorcheck(retcode) < 0:
        return -1

    # event loop
    endofsequence = False
    while True:
        eventtype, data = bufdatacb.queue.get()  # waits until element available
        if eventtype == QUEUE_DATA:
            dataprocessor.process_data(data)
            if endofsequence:
                break
        elif eventtype == QUEUE_ENDOFMEAS:
            retcode = bufdatacb.start_measurement(100)
            if errorcheck(retcode) < 0:
                return -1
        elif eventtype == QUEUE_ENDOFSEQUENCE:
            endofsequence = True
        else: # unknown event
            break # break out of the event loop

    end = timeit.default_timer()
    print("\ntime elapsed : ", end-start, "s")

    time.sleep(0.1)
    # clean up
    bufdatacb.close() # closes the user callbacks pipe, method inherited from base class
    device.deinitialize()

    # show processed data
    print("total events {}".format(dataprocessor.total_events))
    if dataprocessor.xy_histo is not None:
        plt.imshow(dataprocessor.xy_histo)
        plt.show()
    return 0

if __name__ == "__main__":
    test3()
