#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2022 Surface Concept GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

-------------------------------------------------------------------------------

Example showing how to set exposure parameters and how to start a measurement 
with multiple camera frames and how to receive meta information and image data
of the individual frames.

Scroll down to the end of the file and edit the last lines to select whether
to run example_camera_frames_no_image() or example_camera_frames_with_image()
"""

import scTDC
import traceback

def example_camera_frames_no_image():
    """Opens a camera frame pipe and starts one measurement, then reads the meta
    data of each frame and prints it. Doesn't use the image data."""
    FRAMES_COUNT = 5
    EXPOSURE_MICROSECS = 5000
    lib = scTDC.scTDClib()
    camera = scTDC.Camera(inifilepath="tdc_gpx3.ini", autoinit=False, lib=lib)
    v = lib.sc_tdc_config_get_library_version()
    
    print("scTDC library version {}.{}.{}".format(v[0], v[1], v[2]))
    ret, msg = camera.initialize()
    print("initialization result: {} {}".format(ret, msg))
    
    if ret >= 0:
        try:
            camera.set_fanspeed(150)
            camera.set_exposure_and_frames(EXPOSURE_MICROSECS, FRAMES_COUNT)
            camera.set_region_of_interest(0, 1940, 0, 1440)
            print("get ROI", camera.get_region_of_interest())
            pipe = scTDC.CamFramePipe(camera)
            camera.do_measurement(1) # the argument value here is arbitrary
            for i in range(FRAMES_COUNT):
                print(pipe.read(timeout_ms = 1000)[0])
            pipe.close()
        finally:
            camera.deinitialize()

def example_camera_frames_with_image():
    """Opens a camera frame pipe and starts one measurement, then reads the meta
    data of each frame and plots images of the first four frames using 
    matplotlib."""
    from matplotlib import pyplot as plt
    FRAMES_COUNT = 5
    PLOT_IMAGES = 4
    EXPOSURE_MICROSECS = 300
    lib = scTDC.scTDClib()
    camera = scTDC.Camera(inifilepath="tdc_gpx3.ini", autoinit=False, lib=lib)

    v = lib.sc_tdc_config_get_library_version()
    print("scTDC library version {}.{}.{}".format(v[0], v[1], v[2]))

    init_result, msg = camera.initialize()
    print("initialization result: {} {}".format(init_result, msg))
    
    if init_result >= 0:
        try:
            camera.set_fanspeed(150)
            camera.set_exposure_and_frames(EXPOSURE_MICROSECS, FRAMES_COUNT)
            # open pipe for receiving image data and frame meta info
            pipeid, pipe = camera.add_frame_pipe()
            camera.do_measurement(1) # the argument value here is arbitrary
            imgcache = [] # prepare a list for storing images
            for i in range(FRAMES_COUNT):
                if i < PLOT_IMAGES:
                    plt.subplot(2, 2, i+1)
                meta, img = pipe.read(timeout_ms=1000)
                # check for errors
                if isinstance(meta, int):
                    # if the first return value (meta) is an int, img is the
                    # error message
                    print(img)
                    break
                # if this point is reached, reading was successful
                # print contents of meta dictionary
                print(*["{}={}".format(x, y) for x, y in meta.items()])
                if img is not None and i < PLOT_IMAGES:
                    imgcache.append(img.copy())
                    plt.imshow(imgcache[-1], cmap='Greys_r')
                if img is None:
                    # (if the camera is configured to run in blob mode, no image
                    # data is sent)
                    print("no image data was received")

            plt.show()
            pipe.close()
        except:
            traceback.print_exc()
        finally:
            camera.deinitialize()

if __name__ == "__main__":
    #example_camera_frames_no_image()
    example_camera_frames_with_image()